/*
**	CopyPanel.c
**
**	Editing panel for configuration copying
**
**	Copyright  1990-1994 by Olaf `Olsen' Barthel
**		All Rights Reserved
*/

#include "termGlobal.h"

enum	{	GAD_SERIAL=1,GAD_MODEM,GAD_SCREEN,GAD_TERMINAL,GAD_EMULATION,
		GAD_CLIPBOARD,GAD_CAPTURE,GAD_COMMANDS,GAD_MISC,GAD_PATHS,GAD_TRANSFER,
		GAD_TRANSLATION,GAD_KEYMACROS,GAD_CURSORKEYS,GAD_FASTMACROS,GAD_COPYTOALL,
		GAD_DEFAULT,GAD_ALL,GAD_CLEAR,GAD_USE,GAD_CANCEL };

enum	{	COPY_MAIN_CONFIG,COPY_DEFAULTS };

STATIC ULONG	Bits		= 0;

STATIC BYTE	CopyMode	= COPY_MAIN_CONFIG,
		AllFileSettings	= FALSE;

STATIC VOID __regargs
CopyConfig(struct Configuration *MyConfig,BYTE Part)
{
	STATIC LONG TypeMappings[15][2] =
	{
		GAD_SERIAL,		PREF_SERIAL,
		GAD_MODEM,		PREF_MODEM,
		GAD_SCREEN,		PREF_SCREEN,
		GAD_TERMINAL,		PREF_TERMINAL,
		GAD_EMULATION,		PREF_EMULATION,
		GAD_CLIPBOARD,		PREF_CLIP,
		GAD_CAPTURE,		PREF_CAPTURE,
		GAD_COMMANDS,		PREF_COMMAND,
		GAD_MISC,		PREF_MISC,
		GAD_PATHS,		PREF_PATH,
		GAD_TRANSFER,		PREF_TRANSFER,
		GAD_TRANSLATION,	PREF_FILE,
		GAD_KEYMACROS,		PREF_FILE,
		GAD_CURSORKEYS,		PREF_FILE,
		GAD_FASTMACROS,		PREF_FILE
	};

	LONG i,PrefType;
	APTR Data;

	for(i = 0 ; i < 15 ; i++)
	{
		if(Part == TypeMappings[i][0])
		{
			PrefType = TypeMappings[i][1];

			break;
		}
	}

	Data = GetConfigEntry(MyConfig,PrefType);

	if(CopyMode == COPY_MAIN_CONFIG)
	{
		if(CreateConfigEntry(MyConfig,PrefType))
		{
			if(!Data)
				ResetConfigEntry(MyConfig,PrefType,FALSE);

			switch(Part)
			{
				case GAD_SERIAL:

					memcpy(MyConfig -> SerialConfig,Config -> SerialConfig,sizeof(struct SerialSettings));
					break;

				case GAD_MODEM:

					memcpy(MyConfig -> ModemConfig,Config -> ModemConfig,sizeof(struct ModemSettings));
					break;

				case GAD_SCREEN:

					memcpy(MyConfig -> ScreenConfig,Config -> ScreenConfig,sizeof(struct ScreenSettings));
					break;

				case GAD_TERMINAL:

					memcpy(MyConfig -> TerminalConfig,Config -> TerminalConfig,sizeof(struct TerminalSettings));
					break;

				case GAD_EMULATION:

					memcpy(MyConfig -> EmulationConfig,Config -> EmulationConfig,sizeof(struct EmulationSettings));
					break;

				case GAD_CLIPBOARD:

					memcpy(MyConfig -> ClipConfig,Config -> ClipConfig,sizeof(struct ClipSettings));
					break;

				case GAD_CAPTURE:

					memcpy(MyConfig -> CaptureConfig,Config -> CaptureConfig,sizeof(struct CaptureSettings));
					break;

				case GAD_COMMANDS:

					memcpy(MyConfig -> CommandConfig,Config -> CommandConfig,sizeof(struct CommandSettings));
					break;

				case GAD_MISC:

					memcpy(MyConfig -> MiscConfig,Config -> MiscConfig,sizeof(struct MiscSettings));
					break;

				case GAD_PATHS:

					memcpy(MyConfig -> PathConfig,Config -> PathConfig,sizeof(struct PathSettings));
					break;

				case GAD_TRANSFER:

					memcpy(MyConfig -> TransferConfig,Config -> TransferConfig,sizeof(struct TransferSettings));
					break;

				case GAD_TRANSLATION:

					strcpy(MyConfig -> FileConfig -> TranslationFileName,Config -> FileConfig -> TranslationFileName);
					break;

				case GAD_KEYMACROS:

					strcpy(MyConfig -> FileConfig -> MacroFileName,Config -> FileConfig -> MacroFileName);
					break;

				case GAD_CURSORKEYS:

					strcpy(MyConfig -> FileConfig -> CursorFileName,Config -> FileConfig -> CursorFileName);
					break;

				case GAD_FASTMACROS:

					strcpy(MyConfig -> FileConfig -> FastMacroFileName,Config -> FileConfig -> FastMacroFileName);
					break;
			}
		}
	}
	else
	{
		if(Data && (PrefType != PREF_FILE || AllFileSettings))
			DeleteConfigEntry(MyConfig,PrefType);
	}
}

BYTE __regargs
CopyPanel(struct Configuration *MyConfig,BYTE Selective)
{
	BYTE			 MadeChanges = FALSE;
	struct LayoutHandle	*Handle;

	if(Handle = LT_CreateHandleTags(Window -> WScreen,
		LH_LocaleHook,	&LocaleHook,
	TAG_DONE))
	{
		struct Window *PanelWindow;

		LT_New(Handle,
			LA_Type,	VERTICAL_KIND,
		TAG_DONE);
		{
			LT_New(Handle,
				LA_Type,	HORIZONTAL_KIND,
				LA_LabelID,	MSG_V36_0028,
			TAG_DONE);
			{
				LONG i;

				for(i = 0 ; i <= GAD_FASTMACROS - GAD_SERIAL ; i++)
				{
					if(!(i % 5))
					{
						LT_New(Handle,
							LA_Type,	VERTICAL_KIND,
						TAG_DONE);
					}

					LT_New(Handle,
						LA_Type,	CHECKBOX_KIND,
						LA_LabelID,	MSG_COPYPANEL_SERIAL_GAD + i,
						LA_ID,		GAD_SERIAL + i,
						GTCB_Checked,	Bits & (1L << i),
					TAG_DONE);

					if(!((i + 1) % 5))
						LT_EndGroup(Handle);
				}

				LT_EndGroup(Handle);
			}

			LT_New(Handle,
				LA_Type,	VERTICAL_KIND,
			TAG_DONE);
			{
				LT_New(Handle,
					LA_Type,	XBAR_KIND,
				TAG_DONE);

				LT_EndGroup(Handle);
			}

			LT_New(Handle,
				LA_Type,	VERTICAL_KIND,
			TAG_DONE);
			{
				LT_New(Handle,
					LA_Type,	CHECKBOX_KIND,
					LA_LabelID,	MSG_COPYPANEL_TO_ALL_ENTRIES_GAD,
					LA_ID,		GAD_COPYTOALL,
					GTCB_Checked,	!MyConfig,
					GA_Disabled,	!MyConfig,
				TAG_DONE);

				LT_New(Handle,
					LA_Type,	MX_KIND,
					LA_LabelID,	MSG_COPYPANEL_COPY_TYPE_GAD,
					LA_ID,		GAD_DEFAULT,
					LA_BYTE,	&CopyMode,
					LAMX_FirstLabel,MSG_COPYPANEL_COPY_GLOBAL_CONFIG_MSG,
					LAMX_LastLabel,	MSG_COPYPANEL_COPY_GLOBAL_DEFAULTS_MSG,
				TAG_DONE);

				LT_EndGroup(Handle);
			}

			LT_New(Handle,
				LA_Type,	VERTICAL_KIND,
			TAG_DONE);
			{
				LT_New(Handle,
					LA_Type,	XBAR_KIND,
				TAG_DONE);

				LT_EndGroup(Handle);
			}

			LT_New(Handle,
				LA_Type,	HORIZONTAL_KIND,
			TAG_DONE);
			{
				LT_New(Handle,
					LA_Type,	BUTTON_KIND,
					LA_LabelID,	MSG_COPYPANEL_ALL_GAD,
					LA_ID,		GAD_ALL,
				TAG_DONE);

				LT_New(Handle,
					LA_Type,	BUTTON_KIND,
					LA_LabelID,	MSG_COPYPANEL_CLEAR_GAD,
					LA_ID,		GAD_CLEAR,
				TAG_DONE);

				LT_EndGroup(Handle);
			}

			LT_New(Handle,
				LA_Type,	VERTICAL_KIND,
			TAG_DONE);
			{
				LT_New(Handle,LA_Type,XBAR_KIND,LAXB_FullSize,TRUE,TAG_DONE);

				LT_EndGroup(Handle);
			}

			LT_New(Handle,LA_Type,HORIZONTAL_KIND,
				LAGR_SameSize,	TRUE,
				LAGR_Spread,	TRUE,
			TAG_DONE);
			{
				LT_New(Handle,
					LA_Type,	BUTTON_KIND,
					LA_LabelID,	MSG_GLOBAL_USE_GAD,
					LA_ID,		GAD_USE,
					LABT_ReturnKey,	TRUE,
					LABT_ExtraFat,	TRUE,
				TAG_DONE);

				LT_New(Handle,
					LA_Type,	BUTTON_KIND,
					LA_LabelID,	MSG_GLOBAL_CANCEL_GAD,
					LA_ID,		GAD_CANCEL,
					LABT_EscKey,	TRUE,
					LABT_ExtraFat,	TRUE,
				TAG_DONE);

				LT_EndGroup(Handle);
			}

			LT_EndGroup(Handle);
		}

		if(PanelWindow = LT_Layout(Handle,LocaleString(MSG_COPYPANEL_COPY_CONFIG_TXT),NULL,0,0,IDCMP_CLOSEWINDOW,0,
			LAWN_HelpHook,		&GuideHook,
			LAWN_Parent,		Window,
			WA_DepthGadget,		TRUE,
			WA_CloseGadget,		TRUE,
			WA_DragBar,		TRUE,
			WA_RMBTrap,		TRUE,
			WA_Activate,		TRUE,
		TAG_DONE))
		{
			struct IntuiMessage	*Message;
			BOOLEAN			 Done = FALSE;
			ULONG			 MsgClass,
						 MsgQualifier;
			UWORD			 MsgCode;
			struct Gadget		*MsgGadget;

			GuideContext(CONTEXT_COPY);

			PushWindow(PanelWindow);

			LT_ShowWindow(Handle,TRUE);

			do
			{
				if(Wait(PORTMASK(PanelWindow -> UserPort) | SIG_BREAK) & SIG_BREAK)
					break;

				while(Message = (struct IntuiMessage *)GT_GetIMsg(PanelWindow -> UserPort))
				{
					MsgClass	= Message -> Class;
					MsgQualifier	= Message -> Qualifier;
					MsgCode		= Message -> Code;
					MsgGadget	= (struct Gadget *)Message -> IAddress;

					GT_ReplyIMsg(Message);

					LT_HandleInput(Handle,MsgQualifier,&MsgClass,&MsgCode,&MsgGadget);

					if(MsgClass == IDCMP_CLOSEWINDOW)
						Done = TRUE;

					if(MsgClass == IDCMP_GADGETUP)
					{
						LONG i;

						switch(MsgGadget -> GadgetID)
						{
							case GAD_ALL:

								for(i = GAD_SERIAL ; i <= GAD_FASTMACROS ; i++)
									LT_SetAttributes(Handle,i,GTCB_Checked,TRUE,TAG_DONE);

								break;

							case GAD_CLEAR:

								for(i = GAD_SERIAL ; i <= GAD_FASTMACROS ; i++)
									LT_SetAttributes(Handle,i,GTCB_Checked,FALSE,TAG_DONE);

								break;

							case GAD_USE:

								Bits = NULL;

								if(LT_GetAttributes(Handle,GAD_COPYTOALL,TAG_DONE))
								{
									LONG j;

									AllFileSettings = TRUE;

									for(i = GAD_TRANSLATION ; i <= GAD_FASTMACROS ; i++)
										AllFileSettings &= LT_GetAttributes(Handle,i,TAG_DONE);

									if(Selective)
									{
										for(i = 0 ; i <= GAD_FASTMACROS - GAD_SERIAL ; i++)
										{
											if(LT_GetAttributes(Handle,GAD_SERIAL + i,TAG_DONE))
											{
												for(j = 0 ; j < NumPhoneEntries ; j++)
												{
													if(Phonebook[j] -> Count != -1)
														CopyConfig(Phonebook[j] -> Config,GAD_SERIAL + i);
												}

												Bits |= (1L << i);

												MadeChanges = TRUE;
											}
										}
									}
									else
									{
										for(i = 0 ; i <= GAD_FASTMACROS - GAD_SERIAL ; i++)
										{
											if(LT_GetAttributes(Handle,GAD_SERIAL + i,TAG_DONE))
											{
												for(j = 0 ; j < NumPhoneEntries ; j++)
													CopyConfig(Phonebook[j] -> Config,GAD_SERIAL + i);

												Bits |= (1L << i);

												MadeChanges = TRUE;
											}
										}
									}
								}
								else
								{
									if(MyConfig)
									{
										for(i = 0 ; i <= GAD_FASTMACROS - GAD_SERIAL ; i++)
										{
											if(LT_GetAttributes(Handle,GAD_SERIAL + i,TAG_DONE))
											{
												CopyConfig(MyConfig,GAD_SERIAL + i);

												Bits |= (1L << i);

												MadeChanges = TRUE;
											}
										}
									}
								}

								Done = TRUE;
								break;

							case GAD_CANCEL:

								Done = TRUE;
								break;
						}
					}
				}
			}
			while(!Done);

			PopWindow();
		}

		LT_DeleteHandle(Handle);
	}

	return(MadeChanges);
}
